#!/usr/bin/env python3

"""The basic unit of TF-IDF, the keyword.

This class allows you to have stemmed keywords, but still see the original text.
"""

# Standard modules
from __future__ import absolute_import
from collections import namedtuple

# Installed modules
from mezcla import debug
from mezcla import system


Location = namedtuple('Location', ['document', 'start', 'end'])


class DocKeyword(object):

    """Class for maintaining stemmed term and original"""
    
    def __init__(self, text, document=None, start=None, end=None):
        self.locations = set()
        self.text = text
        if (start is not None) and (end is not None):
            self.locations.add(Location(document, start, end))
        ## TODO: debug.trace_object(8, self, "DocKeyword instance", show_all=False)
        debug.trace_object(7, self, "DocKeyword instance", show_all=False)

    def update_locations(self, locations):
        """Add LOCATIONS to other locations"""
        self.locations = self.locations.union(locations)

    def __add__(self, other):
        assert self.text == other.text
        out = DocKeyword(self.text)
        out.locations = self.locations
        out.update_locations(other.locations)
        return out

    def __ladd__(self, other):
        if other == 0:
            return self
        return self.__add__(other)

    def __len__(self):
        return len(self.locations)

    @property
    def original_texts(self):
        """Return list of original texts"""
        out = []
        for loc in self.locations:
            if loc.document:
                text = loc.document.text[loc.start:loc.end]
            else:
                text = ''
            out.append(text)
        return list(set(out))

    def get_first_text(self):
        """Return the first original text."""
        loc = next(iter(self.locations))
        return loc.document.text[loc.start:loc.end]

    def __str__(self):
        return u'Stem:%s, Instances:%s, Count:%d' % (self.text, str(self.original_texts), len(self))

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone")
    debug.trace_object(3, DocKeyword("monkeys"))



#! /usr/bin/env python
#
# Test(s) for ../merge_files.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_merge_files.py
#

"""Tests for merge_files module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.merge_files as THE_MODULE

class TestMergeFiles(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__)
    use_temp_base_dir = True            # treat TEMP_BASE as directory

    def test_get_timestamp(self):
        """Ensure get_timestamp works as expected"""
        debug.trace(4, f"test_get_timestamp(); self={self}")
        # Note: "/sbin should be from April of LTS year (2018, 2022, etc.)
        assert(re.search(r"(2018|2022)-04-\d\d \d\d:\d\d:\d\d", THE_MODULE.get_timestamp("/sbin")))

    def test_get_numeric_timestamp(self):
        """Ensure get_numeric_timestamp works as expected"""
        debug.trace(4, f"test_get_numeric_timestamp(); self={self}")
        # Note: /tmp should be newer than /vmlinuz
        assert(THE_MODULE.get_numeric_timestamp("/vmlinuz") < THE_MODULE.get_numeric_timestamp("/tmp"))

    def test_simple_merge(self):
        """Make sure simple merge works"""
        debug.trace(4, "test_simple_merge()")
        tmp = self.temp_base
        system.write_lines(f"{tmp}/1.list", map(str, [0, 1, 2, 3]))
        gh.full_mkdir(f"{tmp}/backup")
        system.write_lines(f"{tmp}/backup/1.list~", map(str, [1, 2, 3]))
        system.write_lines(f"{tmp}/other-1.list", map(str, [1, 2, 3, 4]))
        assert(self.run_script(env_options="IGNORE_TIMESTAMP=1", uses_stdin=True, data_file="",
                               options=f"{tmp}/1.list {tmp}/other-1.list").split()
               == "0 1 2 3 4".split())

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Based on following:
#   TODO: url
#

"""TODO: overview"""

# Standard modules
# TODO: import re

# Intalled module
# TODO: import numpy as np

# Local modules
from mezcla import debug
from mezcla.main import Main
from mezcla import system
## TODO:
## from mezcla import glue_helpers as gh

# Constants
TL = debug.TL

## TODO:
## # Environment options
## # Notes:
## # - These are just intended for internal options, not for end users.
## # - They also allow for enabling options in one place rather than four
## #   when using main.Main (e.g., [Main member] initialization, run-time
## #   value, and argument spec., along with string constant definition).
## #
## ENABLE_FUBAR = system.getenv_bool("ENABLE_FUBAR", False,
##                                   description="Enable fouled up beyond all recognition processing")

#-------------------------------------------------------------------------------

def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Show simple usage if --help given
    dummy_main_app = Main(description=__doc__, skip_input=False, manual_input=False)
    debug.assertion(dummy_main_app.parsed_args)

    ## TODO:
    system.print_error("Error: Implement me!")
    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python
#
# Test(s) for ../file_utils.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_file_utils.py
#

"""Tests for file_utils module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla import glue_helpers as gh

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.file_utils as THE_MODULE

class TestFileUtils(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.derive_tested_module_name(__file__)
    use_temp_base_dir = True

    ## TODO: optional setup methods
    ##
    ## @classmethod
    ## def setUpClass(cls):
    ##     """One-time initialization (i.e., for entire class)"""
    ##     debug.trace(6, f"TestFileUtils.setUpClass(); cls={cls}")
    ##     # note: should do parent processing first
    ##     super().setUpClass()
    ##     ...
    ##     return
    ##
    ## def setUp(self):
    ##     """Per-test setup"""
    ##     debug.trace(6, f"TestFileUtils.setUp(); self={self}")
    ##     # note: must do parent processing first (e.g., for temp file support)
    ##     super().setUp()
    ##     ...
    ##     return

    def test_get_directory_listing(self):
        """
        Tests for get_directory_listing(path          = 'PATH',
                                        recursive     = False,
                                        long          = False,
                                        readable_size = False,
                                        return_string = True,
                                        make_unicode  = False)
        """

        # Setup files and folders
        folders     = [self.temp_base, f'{self.temp_base}/other_folder']
        filenames   = ['analize.py', 'debug.cpp', 'main.txt', 'misc_utils.perl']

        for foldername in folders:
            gh.run(f'mkdir {foldername}')
            for filename in filenames:
                gh.run(f'touch {foldername}/{filename}')

        # Run Test
        list_result = THE_MODULE.get_directory_listing(f'{self.temp_base}', recursive=True, long=True, return_string=True)

        for line in list_result:
            assert bool(re.search(r"[drwx-]+\s+\d+\s+\w+\s+\w+\s+\d+\s+\w+\s+\d+\s+\d\d:\d\d\s+[\w/]+", line))

    def test_get_information(self):
        """Tests for def get_information(path,
                                         readable = False,
                                         return_string = False)
        """
        gh.run(f'touch {self.temp_file}')

        ls_result = gh.run(f'ls -l {self.temp_file}')
        ls_result = re.sub(r'\s+', ' ', ls_result)

        assert THE_MODULE.get_information(self.temp_file, return_string=True).lower() == ls_result.lower()

    def test_get_permissions(self):
        """Tests for get_permissions(path)"""

        # Setup
        test_file = self.temp_base + '/' + 'some-file.cpp'
        gh.run(f'touch {test_file}')

        # Run
        assert THE_MODULE.get_permissions(test_file) == gh.run(f'ls -l {test_file}')[:10]
        assert THE_MODULE.get_permissions(self.temp_base) == gh.run(f'ls -ld {self.temp_base}')[:10]

    def test_get_modification_date(self):
        """Tests for get_modification_date(path)"""
        gh.run(f'touch {self.temp_file}')

        ls_date = gh.run(f'ls -l {self.temp_file}').lower()
        ls_date = re.search(r'\w\w\w +\d\d +\d\d:\d\d', ls_date).group()

        assert THE_MODULE.get_modification_date(self.temp_file, strftime='%b %-d %H:%M').lower() == ls_date

    ## TODO: optional cleanup methods
    ##
    ## def tearDown(self):
    ##     debug.trace(6, f"TestFileUtils.tearDown(); self={self}")
    ##     super(TestFileUtils, cls).tearDownClass()
    ##     ...
    ##     return
    ##
    ## @classmethod
    ## def tearDownClass(cls):
    ##     debug.trace(6, f"TestFileUtils.tearDownClass(); cls={cls}")
    ##     super(TestFileUtils, self).tearDown()
    ##     ...
    ##     return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# spell.py: simple spell checker (e.g., via Enchant module)
#
# TODO:
# - integrate code for suggestions based on spelling distance
#

"""Basic spell checking"""

# Standard modules
# TODO: sys => system
import sys
import fileinput

# Installed modules
import enchant			# spell checking

# Local modules
from mezcla import debug
## TODO: from mezcla.main import Main
## TODO: from mezcla import system
## OLD: from mezcla.my_regex import my_re
from mezcla.text_processing import split_word_tokens

## TEMP:
# pylint: disable=consider-using-f-string

# Process command line
# TODO: upgrade to using Main script (see template.py)
i = 1
show_usage = (i == len(sys.argv))
while (i < len(sys.argv)) and (sys.argv[i][0] == "-"):
    if (sys.argv[i] == "--help"):
        show_usage = True
    elif (sys.argv[i] == "-"):
        pass
    else:
        print("Error: unexpected argument '%s'" % sys.argv[i])
        show_usage = True
    i += 1
if (show_usage):
    print("Usage: %s [options] input-file" % sys.argv[0])
    print("")
    print("Options: [--help]")
    print("")
    print("Examples:")
    print("")
    print("  %s query-keywords.list" % sys.argv[0])
    print("")
    print("  echo 'how now browne cow?' | {prog} -".format(prog=sys.argv[0]))
    print("")
    sys.exit()
# Discard any used arguments (for sake of fileinput)
if (i > 1):
    debug.trace(5, f"discarding used args: {sys.argv[1: i]}")
    sys.argv = [sys.argv[0]] + sys.argv[i:]

# Initialize spell checking
speller = enchant.Dict("en_US")

# Check input
for line in fileinput.input():
    line = line.strip()
    debug.trace_fmt(5, "L{line_num}: {line_text}", line_num=fileinput.filelineno(), line_text=line)

    # Extract word tokens and print those not recognized
    ## BAD:
    ## word_tokens = [t.strip() for t in my_re.split(r"\W+", line.lower(), my_re.LOCALE|my_re.UNICODE)
    ##                if (len(t.strip()) > 0)]
    word_tokens = split_word_tokens(line.lower())
    debug.trace(4, f"tokens: {word_tokens}")
    for w in word_tokens:
        if not speller.check(w):
            print(w)



#! /usr/bin/env python
#
# Displays names of web colors for RGB codes, using a distance match in
# case no exact matches exists.
#
# Notes:
# - Based on https://computergraphics.stackexchange.com/questions/1542/convert-rgb-hex-or-any-other-color-format-to-standard-color-programmatically.
# - Color rgb:(0, 255, 0) is reported as lime instead of green. See
#     https://webcolors.readthedocs.io
# - For utility to extract colors from images, see
#   https://pypi.org/project/extcolors
#
#--------------------------------------------------------------------------------
# Sample input (based on extcolors):
#
#   Extracted colors:
#   (128, 128, 128):  72.98% (1888)
#   (39, 39, 39)   :  24.35% (630)
#   (255, 255, 255):   2.67% (69)
#   
#   Pixels in output: 2587 of 11648
#
# Sample output:
#
#   Extracted colors:
#   <(128, 128, 128), gray>:  72.98% (1888)
#   <(39, 39, 39), darkslategray>   :  24.35% (630)
#   <(255, 255, 255), white>:   2.67% (69)
#   
#   Pixels in output: 2587 of 11648
#
#................................................................................
# TODO:
# - Add example usage with extcolors:
#   extcolors self.filename | self.program -
# - Add option to invoke extcolors automatically.
#

"""Convert RGB tuples into color names"""

# Standard packages
import re

# Installed packages
import webcolors
from scipy.spatial import KDTree

# Local packages
from mezcla import debug
from mezcla.main import Main
from mezcla import system
from mezcla.my_regex import my_re

# Constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")

## TODO: REPLACEMENT = "regex-replacement"
RGB_REGEX = "rgb-regex"
HEX = "hex"
SKIP_DIRECT = "skip-direct"

class Script(Main):
    """Input processing class: convert RGB tuples to <RGB, label> pairs"""
    rgb_regex = r"\((0?x?[0-9A-F]+), (0?x?[0-9A-F]+), (0?x?[0-9A-F]+)\)"
    ## TODO: replacement = r"<COLOR, \1>"
    space_color_db = None
    color_names = []
    hex = False
    skip_direct = False
    check_direct_match = None

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        # Extract argument values
        ## TODO: self.REPLACEMENT = self.get_parsed_option(REPLACEMENT, self.REPLACEMENT)
        self.rgb_regex = self.get_parsed_option(RGB_REGEX, self.rgb_regex)
        spec_regex = r"\([^\(\)]+\).*" * 3
        debug.assertion(re.search(spec_regex, self.rgb_regex, re.IGNORECASE))
        self.hex = self.get_parsed_option(HEX, self.hex)
        self.skip_direct = self.get_parsed_option(SKIP_DIRECT, self.skip_direct)
        self.check_direct_match = not self.skip_direct

        # Populate color names into spatial name database
        hexnames = webcolors.CSS3_HEX_TO_NAMES
        self.color_names = []
        color_positions = []
        #
        for hex_code, name in hexnames.items():
            debug.trace(6, f"color: {name}={hex_code}")
            self.color_names.append(name)
            color_positions.append(webcolors.hex_to_rgb(hex_code))
            #
        self.space_color_db = KDTree(color_positions)
        debug.trace_object(5, self, label="Script instance")

    def process_line(self, line):
        """Processes current line from input"""
        debug.trace_fmtd(6, "Script.process_line({l})", l=line)

        # Do sanity check for inadvertant image input
        # TODO: abort processing unless --force option given
        if (self.line_num <= 1):
            debug.assertion(not re.search(r"^\s*(IMDR|JFIF|PNG)\s*$", line),
                            "Input should not be an image (e.g., use extcolors output)")
        
        # Extract RGB references and add color name label
        # ex: "(128, 128, 128):  72.98% (1888)" => "<Grey, (128, 128, 128)>:  72.98% (1888)
        ## OLD: MAX_TRIES = max(1, line.count("("))
        MAX_TRIES = (1 + len(re.findall(self.rgb_regex, line)))
        debug.trace(5, f"len: {len(line)}; MAX_TRIES={MAX_TRIES}")
        num_tries = 0
        text = line
        processed_text = ""
        while (my_re.search(self.rgb_regex, text) and (num_tries < MAX_TRIES)):
            num_tries += 1
            # Extract RGB components
            rgb = my_re.group(0)
            red = my_re.group(1)
            green = my_re.group(2)
            blue = my_re.group(3)

            # Determine whether RGB in hexadecimal or decimal
            rgb_base = 10
            if (self.hex or re.search("(0x)|[A-F]|(^#)", rgb, re.IGNORECASE)):
                rgb_base = 16
            query_color = [system.safe_int(c, base=rgb_base) for c in [red, green, blue]]

            try:
                # Try for exact match
                color_name = None
                if self.check_direct_match:
                    try:
                        color_name = webcolors.rgb_to_name(query_color)
                    except:
                        debug.trace(5, f"Direct lookup failed: {system.get_exception()}")
                    
                # Query nearest point
                if not color_name:
                    dist, index = self.space_color_db.query(query_color)
                    debug.trace_fmtd(5, f"{rgb} => {index}; dist={dist}")
                    color_name = self.color_names[index]
            except:
                system.print_stderr(f"Exception in color decoding: {system.get_exception()}")
                continue

            rgb_spec = rgb
            color_spec = f"<{rgb_spec}, {color_name}>"
            processed_text += text[0: my_re.start()] + color_spec
            text = text[my_re.end():]
            debug.trace_fmtd(4, "match: {m}; new text: {new}",
                             m=my_re.group(0), new=text)
        debug.assertion(num_tries <= MAX_TRIES)

        # Print revised line
        print(processed_text + text)

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    app = Script(
        description=__doc__,
        usage_notes="Note: Input is not an image (e.g., use extcolors)",
        # Note: skip_input controls the line-by-line processing, which is inefficient but simple to
        # understand; in contrast, manual_input controls iterator-based input (the opposite of both).
        skip_input=False,
        manual_input=False,
        boolean_options=[(HEX, "RGB triple specified in hex (not decimal)"),
                         (SKIP_DIRECT, "Don't include direct match (for nearest neighbor test")],
        # Note: FILENAME is default argument unless skip_input
        text_options=[
            ## TODO: (REPLACEMENT, "Regex-like replacement using \\1 for RGB tuple and COLOR for color name"),
            (RGB_REGEX, "Regex for finding RGB color specifications with \\1, \\2, and \\3 for reg, green, and blue")
        ],
        # Note: Following added for indentation: float options are not common
        float_options=None)
    app.run()
